/***********************************************************
*                      picoQuadcopter                      *
*                                                          *
*                        calc flie                         *
*   (from "bitcraze-crazyflie-firmware-669962cac073.zip"   *
***********************************************************/
#include "../config.h"

#include <math.h>
#include <stdint.h>

// ***************** from "bitcraze-crazyflie-firmware-669962cac073.zip" ********************
#define IMU_UPDATE_DT                (float)(1.0 / IMU_UPDATE_FREQ)

#define ATTITUDE_UPDATE_RATE_DIVIDER  2
#define FUSION_UPDATE_DT             (float)(1.0 / (IMU_UPDATE_FREQ / ATTITUDE_UPDATE_RATE_DIVIDER))

#define IIR_SHIFT                     8 
#define IMU_ACC_WANTED_LPF_CUTOFF_HZ  10
#define IMU_ACC_IIR_LPF_ATTENUATION  (IMU_UPDATE_FREQ / (2 * 3.1415 * IMU_ACC_WANTED_LPF_CUTOFF_HZ))
#define IMU_ACC_IIR_LPF_ATT_FACTOR   (int)(((1 << IIR_SHIFT) / IMU_ACC_IIR_LPF_ATTENUATION) + 0.5)

//Fancier version
#define TRUNCATE_SINT16(out, in) (out = (S16)((in<INT16_MIN)?INT16_MIN:((in>INT16_MAX)?INT16_MAX:in)))
//Better semantic
#define SATURATE_SINT16(in) ( (in<INT16_MIN)?INT16_MIN:((in>INT16_MAX)?INT16_MAX:in) )


#define M_PI                  3.14159265358979323846

typedef struct {
    S16 x;
    S16 y;
    S16 z;
} Axis3i16;

typedef struct {
    S32 x;
    S32 y;
    S32 z;
} Axis3i32;

typedef struct {
    float x;
    float y;
    float z;
} Axis3f;

typedef struct
{
    float desired;      //< set point
    float error;        //< error
    float prevError;    //< previous error
    float integ;        //< integral
    float deriv;        //< derivative
    float kp;           //< proportional gain
    float ki;           //< integral gain
    float kd;           //< derivative gain
    float outP;         //< proportional output (debugging)
    float outI;         //< integral output (debugging)
    float outD;         //< derivative output (debugging)
    float iLimit;       //< integral limit
} PidObject;

float eulerRollActual;
float eulerPitchActual;
float eulerYawActual;
float eulerRollDesired;
float eulerPitchDesired;
float eulerYawDesired;
float rollRateDesired;
float pitchRateDesired;
float yawRateDesired;
float fusionDt;

//                        des               kp    ki    kd            limit
PidObject pidRollRate  = {0.0, 0, 0, 0, 0, 70.0,  0.0,  0.0, 0, 0, 0, 100.0};
PidObject pidPitchRate = {0.0, 0, 0, 0, 0, 70.0,  0.0,  0.0, 0, 0, 0, 100.0};
PidObject pidYawRate   = {0.0, 0, 0, 0, 0, 50.0, 25.0,  0.0, 0, 0, 0, 500.0};
PidObject pidRoll      = {0.0, 0, 0, 0, 0,  3.5,  2.0,  0.0, 0, 0, 0,  20.0};
PidObject pidPitch     = {0.0, 0, 0, 0, 0,  3.5,  2.0,  0.0, 0, 0, 0,  20.0};
PidObject pidYaw       = {0.0, 0, 0, 0, 0,  0.0,  0.0,  0.0, 0, 0, 0, 360.0};

S16  rollOutput;
S16  pitchOutput;
S16  yawOutput;

U16  actuatorThrust;
S16  actuatorRoll;
S16  actuatorPitch;
S16  actuatorYaw;




#define MADWICK_QUATERNION_IMU

#ifdef MADWICK_QUATERNION_IMU
  #define BETA_DEF     0.01f    // 2 * proportional gain
#else // MAHONY_QUATERNION_IMU
    #define TWO_KP_DEF  (2.0f * 0.4f) // 2 * proportional gain
    #define TWO_KI_DEF  (2.0f * 0.001f) // 2 * integral gain
#endif

#ifdef MADWICK_QUATERNION_IMU
  float beta = BETA_DEF;     // 2 * proportional gain (Kp)
#else // MAHONY_QUATERNION_IMU
  float twoKp = TWO_KP_DEF;    // 2 * proportional gain (Kp)
  float twoKi = TWO_KI_DEF;    // 2 * integral gain (Ki)
  float integralFBx = 0.0f;
  float integralFBy = 0.0f;
  float integralFBz = 0.0f;  // integral error terms scaled by Ki
#endif

float q0 = 1.0f;
float q1 = 0.0f;
float q2 = 0.0f;
float q3 = 0.0f;  // quaternion of sensor frame relative to auxiliary frame


//---------------------------------------------------------------------------------------------------
// Fast inverse square-root
// See: http://en.wikipedia.org/wiki/Fast_inverse_square_root
float invSqrt(float x)
{
    float halfx = 0.5f * x;
    float y = x;
    long i = *(long*)&y;
    i = 0x5f3759df - (i>>1);
    y = *(float*)&i;
    y = y * (1.5f - (halfx * y * y));
    return y;
}

float pidUpdate(PidObject* pid, const float measured, const BOOLEAN updateError)
{
    float output;

    if (updateError) {
        pid->error = pid->desired - measured;
    }
    pid->integ += pid->error * IMU_UPDATE_DT;
    if (pid->integ > pid->iLimit) {
        pid->integ = pid->iLimit;
    } else if (pid->integ < -pid->iLimit) { 
        pid->integ = -pid->iLimit;
    }
    pid->deriv = (pid->error - pid->prevError) / IMU_UPDATE_DT;
    pid->outP = pid->kp * pid->error;
    pid->outI = pid->ki * pid->integ;
    pid->outD = pid->kd * pid->deriv;
    output = (pid->kp * pid->error) + (pid->ki * pid->integ) + (pid->kd * pid->deriv);
    pid->prevError = pid->error;
    return output;
}

void controllerCorrectRatePID(
       float rollRateActual, float pitchRateActual, float yawRateActual,
       float rollRateDesired, float pitchRateDesired, float yawRateDesired)
{
    pidRollRate.desired = rollRateDesired;
    TRUNCATE_SINT16(rollOutput, pidUpdate(&pidRollRate, rollRateActual, TRUE));
    pidPitchRate.desired = pitchRateDesired;
    TRUNCATE_SINT16(pitchOutput, pidUpdate(&pidPitchRate, pitchRateActual, TRUE));
    pidYawRate.desired = yawRateDesired;
    TRUNCATE_SINT16(yawOutput, pidUpdate(&pidYawRate, yawRateActual, TRUE));
}

void controllerCorrectAttitudePID(
       float eulerRollActual, float eulerPitchActual, float eulerYawActual,
       float eulerRollDesired, float eulerPitchDesired, float eulerYawDesired,
       float* rollRateDesired, float* pitchRateDesired, float* yawRateDesired)
{
    pidRoll.desired = eulerRollDesired;
    *rollRateDesired = pidUpdate(&pidRoll, eulerRollActual, TRUE);
    // Update PID for pitch axis
    pidPitch.desired = eulerPitchDesired;
    *pitchRateDesired = pidUpdate(&pidPitch, eulerPitchActual, TRUE);
    // Update PID for yaw axis
    {
        float yawError;
        yawError = eulerYawDesired - eulerYawActual;
        if (yawError > 180.0)
            yawError -= 360.0;
        else if (yawError < -180.0)
            yawError += 360.0;
        pidYaw.error = yawError;
     }
     *yawRateDesired = pidUpdate(&pidYaw, eulerYawActual, FALSE);
}

void pidReset(PidObject* pid)
{
    pid->error = 0;
    pid->prevError = 0;
    pid->integ = 0;
    pid->deriv = 0;
}

void controllerResetAllPID(void)
{
    pidReset(&pidRoll);
    pidReset(&pidPitch);
    pidReset(&pidYaw);
    pidReset(&pidRollRate);
    pidReset(&pidPitchRate);
    pidReset(&pidYawRate);
}

void controllerGetActuatorOutput(int16_t* roll, int16_t* pitch, int16_t* yaw)
{
     *roll  = rollOutput;
     *pitch = pitchOutput;
     *yaw   = yawOutput;
}

#ifdef MADWICK_QUATERNION_IMU
// Implementation of Madgwick's IMU and AHRS algorithms.
// See: http://www.x-io.co.uk/open-source-ahrs-with-x-imu
//
// Date     Author          Notes
// 29/09/2011 SOH Madgwick    Initial release
// 02/10/2011 SOH Madgwick  Optimised for reduced CPU load
void sensfusion6UpdateQ(float gx, float gy, float gz, float ax, float ay, float az, float dt)
{
  float recipNorm;
  float s0, s1, s2, s3;
  float qDot1, qDot2, qDot3, qDot4;
  float _2q0, _2q1, _2q2, _2q3, _4q0, _4q1, _4q2 ,_8q1, _8q2, q0q0, q1q1, q2q2, q3q3;

  // Rate of change of quaternion from gyroscope
  qDot1 = 0.5f * (-q1 * gx - q2 * gy - q3 * gz);
  qDot2 = 0.5f * (q0 * gx + q2 * gz - q3 * gy);
  qDot3 = 0.5f * (q0 * gy - q1 * gz + q3 * gx);
  qDot4 = 0.5f * (q0 * gz + q1 * gy - q2 * gx);

  // Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
  if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f)))
  {
    // Normalise accelerometer measurement
    recipNorm = invSqrt(ax * ax + ay * ay + az * az);
    ax *= recipNorm;
    ay *= recipNorm;
    az *= recipNorm;

    // Auxiliary variables to avoid repeated arithmetic
    _2q0 = 2.0f * q0;
    _2q1 = 2.0f * q1;
    _2q2 = 2.0f * q2;
    _2q3 = 2.0f * q3;
    _4q0 = 4.0f * q0;
    _4q1 = 4.0f * q1;
    _4q2 = 4.0f * q2;
    _8q1 = 8.0f * q1;
    _8q2 = 8.0f * q2;
    q0q0 = q0 * q0;
    q1q1 = q1 * q1;
    q2q2 = q2 * q2;
    q3q3 = q3 * q3;

    // Gradient decent algorithm corrective step
    s0 = _4q0 * q2q2 + _2q2 * ax + _4q0 * q1q1 - _2q1 * ay;
    s1 = _4q1 * q3q3 - _2q3 * ax + 4.0f * q0q0 * q1 - _2q0 * ay - _4q1 + _8q1 * q1q1 + _8q1 * q2q2 + _4q1 * az;
    s2 = 4.0f * q0q0 * q2 + _2q0 * ax + _4q2 * q3q3 - _2q3 * ay - _4q2 + _8q2 * q1q1 + _8q2 * q2q2 + _4q2 * az;
    s3 = 4.0f * q1q1 * q3 - _2q1 * ax + 4.0f * q2q2 * q3 - _2q2 * ay;
    recipNorm = invSqrt(s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
    s0 *= recipNorm;
    s1 *= recipNorm;
    s2 *= recipNorm;
    s3 *= recipNorm;

    // Apply feedback step
    qDot1 -= beta * s0;
    qDot2 -= beta * s1;
    qDot3 -= beta * s2;
    qDot4 -= beta * s3;
  }

  // Integrate rate of change of quaternion to yield quaternion
  q0 += qDot1 * dt;
  q1 += qDot2 * dt;
  q2 += qDot3 * dt;
  q3 += qDot4 * dt;

  // Normalise quaternion
  recipNorm = invSqrt(q0*q0 + q1*q1 + q2*q2 + q3*q3);
  q0 *= recipNorm;
  q1 *= recipNorm;
  q2 *= recipNorm;
  q3 *= recipNorm;
}

#else // MAHONY_QUATERNION_IMU

// Madgwick's implementation of Mayhony's AHRS algorithm.
// See: http://www.x-io.co.uk/open-source-ahrs-with-x-imu
//
// Date     Author      Notes
// 29/09/2011 SOH Madgwick    Initial release
// 02/10/2011 SOH Madgwick  Optimised for reduced CPU load
void sensfusion6UpdateQ(float gx, float gy, float gz, float ax, float ay, float az, float dt)
{
  float recipNorm;
  float halfvx, halfvy, halfvz;
  float halfex, halfey, halfez;
  float qa, qb, qc;

  gx = gx * M_PI / 180;
  gy = gy * M_PI / 180;
  gz = gz * M_PI / 180;

  // Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
  if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f)))
  {
    // Normalise accelerometer measurement
    recipNorm = invSqrt(ax * ax + ay * ay + az * az);
    ax *= recipNorm;
    ay *= recipNorm;
    az *= recipNorm;

    // Estimated direction of gravity and vector perpendicular to magnetic flux
    halfvx = q1 * q3 - q0 * q2;
    halfvy = q0 * q1 + q2 * q3;
    halfvz = q0 * q0 - 0.5f + q3 * q3;

    // Error is sum of cross product between estimated and measured direction of gravity
    halfex = (ay * halfvz - az * halfvy);
    halfey = (az * halfvx - ax * halfvz);
    halfez = (ax * halfvy - ay * halfvx);

    // Compute and apply integral feedback if enabled
    if(twoKi > 0.0f)
    {
      integralFBx += twoKi * halfex * dt;  // integral error scaled by Ki
      integralFBy += twoKi * halfey * dt;
      integralFBz += twoKi * halfez * dt;
      gx += integralFBx;  // apply integral feedback
      gy += integralFBy;
      gz += integralFBz;
    }
    else
    {
      integralFBx = 0.0f; // prevent integral windup
      integralFBy = 0.0f;
      integralFBz = 0.0f;
    }

    // Apply proportional feedback
    gx += twoKp * halfex;
    gy += twoKp * halfey;
    gz += twoKp * halfez;
  }

  // Integrate rate of change of quaternion
  gx *= (0.5f * dt);   // pre-multiply common factors
  gy *= (0.5f * dt);
  gz *= (0.5f * dt);
  qa = q0;
  qb = q1;
  qc = q2;
  q0 += (-qb * gx - qc * gy - q3 * gz);
  q1 += (qa * gx + qc * gz - q3 * gy);
  q2 += (qa * gy - qb * gz + q3 * gx);
  q3 += (qa * gz + qb * gy - qc * gx);

  // Normalise quaternion
  recipNorm = invSqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
  q0 *= recipNorm;
  q1 *= recipNorm;
  q2 *= recipNorm;
  q3 *= recipNorm;
}
#endif

void sensfusion6GetEulerRPY(float* roll, float* pitch, float* yaw)
{
    float gx, gy, gz; // estimated gravity direction

    gx = 2 * (q1*q3 - q0*q2);
    gy = 2 * (q0*q1 + q2*q3);
    gz = q0*q0 - q1*q1 - q2*q2 + q3*q3;

    *yaw = atan2(2*q1*q2 - 2*q0*q3, 2*q0*q0 + 2*q1*q1 - 1) * 180 / M_PI;
    *pitch = atan(gx / sqrt(gy*gy + gz*gz)) * 180 / M_PI;
    *roll = atan(gy / sqrt(gx*gx + gz*gz)) * 180 / M_PI;
}


#define cosPitch 1
#define sinPitch 0
#define cosRoll  1
#define sinRoll  0
static void imuAccAlignToGravity(Axis3i16* in, Axis3i16* out)
{
  Axis3i16 rx;
  Axis3i16 ry;

  // Rotate around x-axis
  rx.x = in->x;
  rx.y = in->y * cosRoll - in->z * sinRoll;
  rx.z = in->y * sinRoll + in->z * cosRoll;

  // Rotate around y-axis
  ry.x = rx.x * cosPitch - rx.z * sinPitch;
  ry.y = rx.y;
  ry.z = -rx.x * sinPitch + rx.z * cosPitch;

  out->x = ry.x;
  out->y = ry.y;
  out->z = ry.z;
}

int16_t iirLPFilterSingle(int32_t in, int32_t attenuation,  int32_t* filt)
{
    int32_t inScaled;
    int32_t filttmp = *filt;
    int16_t out;

    if (attenuation > (1<<IIR_SHIFT)) {
        attenuation = (1<<IIR_SHIFT);
    } else if (attenuation < 1) {
        attenuation = 1;
    }
    // Shift to keep accuracy
    inScaled = in << IIR_SHIFT;
    // Calculate IIR filter
    filttmp = filttmp + (((inScaled-filttmp) >> IIR_SHIFT) * attenuation);
    // Scale and round
    out = (filttmp >> 8) + ((filttmp & (1 << (IIR_SHIFT - 1))) >> (IIR_SHIFT - 1));
    *filt = filttmp;
    return out;
} 

static void imuAccIIRLPFilter(Axis3i16* in, Axis3i16* out, Axis3i32* storedValues, int32_t attenuation)
{
    out->x = iirLPFilterSingle(in->x, attenuation, &storedValues->x);
    out->y = iirLPFilterSingle(in->y, attenuation, &storedValues->y);
    out->z = iirLPFilterSingle(in->z, attenuation, &storedValues->z);
}



U32      attitudeCounter = 0;
Axis3i32 accelStoredFilterValues = {0,0,0};
#define  imuAccLpfAttFactor IMU_ACC_IIR_LPF_ATT_FACTOR

void crasyflie_calc(C_VAL *curr, ACT *act)
{
    Axis3i16 imu_acc, accelLPFAligned;
    Axis3f   g, a;

    imuAccIIRLPFilter((Axis3i16 *)&curr->acc_x, &imu_acc, &accelStoredFilterValues, (int32_t)imuAccLpfAttFactor);

    imuAccAlignToGravity(&imu_acc, &accelLPFAligned);
    
    g.x = curr->gyr_pitch   * IMU_DEG_PER_LSB;
    g.y = curr->gyr_roll    * IMU_DEG_PER_LSB;
    g.z = curr->gyr_yaw     * IMU_DEG_PER_LSB;
    a.x = accelLPFAligned.x * IMU_G_PER_LSB;
    a.y = accelLPFAligned.y * IMU_G_PER_LSB;
    a.z = accelLPFAligned.z * IMU_G_PER_LSB;
    
//    commanderGetRPY(&eulerRollDesired, &eulerPitchDesired, &eulerYawDesired);
    eulerRollDesired  = curr->usr_roll;
    eulerPitchDesired = curr->usr_pitch;
    eulerYawDesired   = curr->usr_yaw;
	
//    commanderGetRPYType(&rollType, &pitchType, &yawType);

    if (++attitudeCounter >= ATTITUDE_UPDATE_RATE_DIVIDER) {
      
        sensfusion6UpdateQ(g.x, g.y, g.z, a.x, a.y, a.z, FUSION_UPDATE_DT);
    
        sensfusion6GetEulerRPY(&eulerRollActual, &eulerPitchActual, &eulerYawActual);

        controllerCorrectAttitudePID(eulerRollActual, eulerPitchActual, eulerYawActual,
                                     eulerRollDesired, eulerPitchDesired, -eulerYawDesired,
                                     &rollRateDesired, &pitchRateDesired, &yawRateDesired);
        attitudeCounter = 0;
    }

/*
      if (rollType == RATE)
      {
        rollRateDesired = eulerRollDesired;
      }
      if (pitchType == RATE)
      {
        pitchRateDesired = eulerPitchDesired;
      }
      if (yawType == RATE)
      {
        yawRateDesired = -eulerYawDesired;
      }
*/
    
    // TODO: Investigate possibility to subtract gyro drift.
    controllerCorrectRatePID(curr->gyr_pitch, -curr->gyr_roll, curr->gyr_yaw,
                             rollRateDesired, pitchRateDesired, yawRateDesired);

    controllerGetActuatorOutput(&act->roll, &act->pitch, &act->yaw);
}


/**********************************************************************************
*            calc with MotionTracking & user-command for motor control            *
**********************************************************************************/
void calc(C_VAL *curr, ACT *act)
{
    curr->usr_pitch *= 2;
    curr->usr_roll  *= 2;
    curr->usr_yaw   *= 2;
    curr->usr_z     *= 2;

    crasyflie_calc(curr, act);
}

void calc_reset(void)
{
    controllerResetAllPID();
}




