#include "calc.h"

#include <math.h>

#define DT 0.01f // 100Hz

// Q diagonal 3x3 with these elements on diagonal
#define Q1 5.0f
#define Q2 100.0f
#define Q3 0.01f
 
// R diagonal 2x2 with these elements on diagonal
#define R1 1000.0f
#define R2 1000.0f
 
struct _kalman_data
{
    float x1, x2, x3;
    float p11, p12, p13, p21, p22, p23, p31, p32, p33;
    float q1, q2, q3;
    float r1, r2;
};
 
typedef struct _kalman_data kalman_data;

U08 kalman_isinit = 0;

// Setup the kalman data struct
void kalman_init(kalman_data *data)
{
    if (kalman_isinit >=2) return;

	  data->x1 = 0.0f;
	  data->x2 = 0.0f;
	  data->x3 = 0.0f;
 
	  // Init P to diagonal matrix with large values since
	  // the initial state is not known
    data->p11 = 1000.0f;
    data->p12 = 0.0f;
    data->p13 = 0.0f;
    data->p21 = 0.0f;
    data->p22 = 1000.0f;
    data->p23 = 0.0f;
    data->p31 = 0.0f;
    data->p32 = 0.0f;
    data->p33 = 1000.0f;
 
    data->q1 = Q1;
    data->q2 = Q2;
    data->q3 = Q3;
    data->r1 = R1;
    data->r2 = R2;

    kalman_isinit++;
}
 
void kalman_innovate(kalman_data *data, float z1, float z2)
{
	float y1, y2;
	float a, b, c;
	float sDet;
	float s11, s12, s21, s22;
	float k11, k12, k21, k22, k31, k32;
	float p11, p12, p13, p21, p22, p23, p31, p32, p33;
 
	// Step 1
	// x(k) = Fx(k-1) + Bu + w:
	data->x1 = data->x1 + DT*data->x2 - DT*data->x3;
	//x2 = x2;
	//x3 = x3;
 
	// Step 2
	// P = FPF'+Q
	a = data->p11 + data->p21*DT - data->p31*DT;
	b = data->p12 + data->p22*DT - data->p32*DT;
	c = data->p13 + data->p23*DT - data->p33*DT;
	data->p11 = a + b*DT - c*DT + data->q1;
	data->p12 = b;
	data->p13 = c;
	data->p21 = data->p21 + data->p22*DT - data->p23*DT;
	data->p22 = data->p22 + data->q2;
	//p23 = p23;
	data->p31 = data->p31 + data->p32*DT - data->p33*DT;
	//p32 = p32;
	data->p33 = data->p33 + data->q3;
 
	// Step 3
	// y = z(k) - Hx(k)
	y1 = z1-data->x1;
	y2 = z2-data->x2;
 
	// Step 4
	// S = HPT' + R
	s11 = data->p11 + data->r1;
	s12 = data->p12;
	s21 = data->p21;
	s22 = data->p22 + data->r2;
 
	// Step 5
	// K = PH*inv(S)
	sDet = 1/(s11*s22 - s12*s21);
	k11 = (data->p11*s22 - data->p12*s21)*sDet;
	k12 = (data->p12*s11 - data->p11*s12)*sDet;
	k21 = (data->p21*s22 - data->p22*s21)*sDet;
	k22 = (data->p22*s11 - data->p21*s12)*sDet;
	k31 = (data->p31*s22 - data->p32*s21)*sDet;
	k32 = (data->p32*s11 - data->p31*s12)*sDet;
 
	// Step 6
	// x = x + Ky
	data->x1 = data->x1 + k11*y1 + k12*y2;
	data->x2 = data->x2 + k21*y1 + k22*y2;
	data->x3 = data->x3 + k31*y1 + k32*y2;
 
	// Step 7
	// P = (I-KH)P
	p11 = data->p11*(1.0f - k11) - data->p21*k12;
	p12 = data->p12*(1.0f - k11) - data->p22*k12;
	p13 = data->p13*(1.0f - k11) - data->p23*k12;
	p21 = data->p21*(1.0f - k22) - data->p11*k21;
	p22 = data->p22*(1.0f - k22) - data->p12*k21;
	p23 = data->p23*(1.0f - k22) - data->p13*k21;
	p31 = data->p31 - data->p21*k32 - data->p11*k31;
	p32 = data->p32 - data->p22*k32 - data->p12*k31;
	p33 = data->p33 - data->p22*k32 - data->p13*k31;
	data->p11 = p11; data->p12 = p12; data->p13 = p13;
	data->p21 = p21; data->p22 = p22; data->p23 = p23;
	data->p31 = p31; data->p32 = p32; data->p33 = p33;
}

// Structs for containing filter data
kalman_data pitch_data;
kalman_data roll_data;


/**********************************************************************************
*            calc with MotionTracking & user-command for motor control            *
**********************************************************************************/
void calc(C_VAL *curr, ACT *act)
{
    float acc_pitch, acc_roll;
    float pitch, roll;

    kalman_init(&pitch_data);
    kalman_init(&roll_data);

    // Calculate pitch and roll based only on acceleration.
    acc_pitch = atan2(curr->acc_x, -curr->acc_z);
    acc_roll  = -atan2(curr->acc_y, -curr->acc_z);

    // Perform filtering
    kalman_innovate(&pitch_data, acc_pitch, curr->gyr_pitch);
    kalman_innovate(&roll_data, acc_roll, curr->gyr_roll);
 
    // The angle is stored in state 1
    pitch = pitch_data.x1;
    roll = roll_data.x1;

    // !!! here we missing the PID !!!

}



