#include "calc.h"
#include "math.h"


#define M_PI                  3.14159265358979323846

#pragma float64

/**************** from goPiCopter-master *****************/

typedef struct ImuMayhony {
    double              lastUpdate;
    float               sampleFreq;  // half the sample period expressed in seconds
    float               exInt, eyInt, ezInt;
    float               q0, q1, q2, q3;
    float               integralFBx;
    float               integralFBy;
    float               integralFBz;
    float               twoKp;  // 2 * poportional gain
    float               twoKi;  // 2 * integral gain
} IMU; 

IMU ImuMayhony = {
    0, 
    0,
    0, 0, 0,
    1, 0, 0, 0,
    0, 0, 0,
    2.0 * 0.5,
    2.0 * 0.1 
};

/**
* Inverse square root
**/
float invSqrt(float x)
{
    return (float(1.0 / (float64)(sqrt(x))));
} 

/**
* Quaternion implementation of the 'DCM filter' [Mayhony et al].  Incorporates the magnetic distortion
* compensation algorithms from Sebastian Madgwick's filter which eliminates the need for a reference
* direction of flux (bx bz) to be predefined and limits the effect of magnetic distortions to yaw
* axis only.
* 
* @see: http://www.x-io.co.uk/open-source-imu-and-ahrs-algorithms/
**/
void ahrsUpdate(float gx, float gy, float gz, 
                float ax, float ay, float az, 
                float mx, float my, float mz)
{
		float    recipNorm;
		float    hx, hy, bx, bz;
		float    halfx, halfy, halfz;
		float    halfex, halfey, halfez;
		float    qa, qb, qc;
		float    q0q0, q0q1, q0q2, q0q3;
		float    q1q1, q1q2, q1q3;
		float    q2q2, q2q3, q3q3;

    // Auxiliary variables to avoid repeated arithmetic
    q0q0 = ImuMayhony.q0 * ImuMayhony.q0;
    q0q1 = ImuMayhony.q0 * ImuMayhony.q1;
    q0q2 = ImuMayhony.q0 * ImuMayhony.q2;
    q0q3 = ImuMayhony.q0 * ImuMayhony.q3;
    q1q1 = ImuMayhony.q1 * ImuMayhony.q1;
    q1q2 = ImuMayhony.q1 * ImuMayhony.q2;
    q1q3 = ImuMayhony.q1 * ImuMayhony.q3;
    q2q2 = ImuMayhony.q2 * ImuMayhony.q2;
    q2q3 = ImuMayhony.q2 * ImuMayhony.q3;
    q3q3 = ImuMayhony.q3 * ImuMayhony.q3;

    // Use magnetometer measurement only when valid (avoids NaN in magnetometer normalisation)
    if (mx != 0.0 && my != 0.0 && mz != 0.0) {
        // Normalize magnetometer measurement
        recipNorm = invSqrt(mx*mx + my*my + mz*mz);
        mx *= recipNorm;
        my *= recipNorm;
        mz *= recipNorm;

        // Reference direction of Earth's magnetic field
        hx = 2.0 * (mx*(0.5-q2q2-q3q3) + my*(q1q2-q0q3) + mz*(q1q3+q0q2));
        hy = 2.0 * (mx*(q1q2+q0q3) + my*(0.5-q1q1-q3q3) + mz*(q2q3-q0q1));
        bx = (float)(sqrt((float64)(hx*hx + hy*hy)));
        bz = 2.0 * (mx*(q1q3-q0q2) + my*(q2q3+q0q1) + mz*(0.5-q1q1-q2q2));

        // Estimated direction of magnetic field
        halfx = bx*(0.5-q2q2-q3q3) + bz*(q1q3-q0q2);
        halfy = bx*(q1q2-q0q3) + bz*(q0q1+q2q3);
        halfz = bx*(q0q2+q1q3) + bz*(0.5-q1q1-q2q2);

        // Error is sum of cross product between estimated direction and measured direction of field vectors
        halfex = (my*halfz - mz*halfy);
        halfey = (mz*halfx - mx*halfz);
        halfez = (mx*halfy - my*halfx);
    }

    // Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
    if (ax != 0.0 && ay != 0.0 && az != 0.0) {
        // Normalize accelerometer measurement
        recipNorm = invSqrt(ax*ax + ay*ay + az*az);
        ax *= recipNorm;
        ay *= recipNorm;
        az *= recipNorm;

        // Estimated direction of gravity
        halfx = q1q3 - q0q2;
        halfy = q0q1 + q2q3;
        halfz = q0q0 - 0.5 + q3q3;

        // Error is sum of cross product between estimated direction and measured direction of field vectors
        halfex += (ay*halfz - az*halfy);
        halfey += (az*halfx - ax*halfz);
        halfez += (ax*halfy - ay*halfx);
    }

    // Apply feedback only when valid data has been gathered from the accelerometer or magnetometer
    if ((halfex != 0.0) && (halfey != 0.0) && (halfez != 0.0)) {
        // Compute and apply integral feedback if enabled
        if (ImuMayhony.twoKi > 0.0) {
            ImuMayhony.integralFBx += ImuMayhony.twoKi * halfex * (1.0 / ImuMayhony.sampleFreq); // integral error scaled by Ki
            ImuMayhony.integralFBy += ImuMayhony.twoKi * halfey * (1.0 / ImuMayhony.sampleFreq);
            ImuMayhony.integralFBz += ImuMayhony.twoKi * halfez * (1.0 / ImuMayhony.sampleFreq);
            gx += ImuMayhony.integralFBx; // apply integral feedback
            gy += ImuMayhony.integralFBy;
            gz += ImuMayhony.integralFBz;
        } else {
            ImuMayhony.integralFBx = 0.0; // prevent integral windup
            ImuMayhony.integralFBy = 0.0;
            ImuMayhony.integralFBz = 0.0;
        }
        // Apply proportional feedback
        gx += ImuMayhony.twoKp * halfex;
        gy += ImuMayhony.twoKp * halfey;
        gz += ImuMayhony.twoKp * halfez;
    }
    // Integrate rate of change of quaternion
    gx *= (0.5 * (1.0 / ImuMayhony.sampleFreq)); // pre-multiply common factors
    gy *= (0.5 * (1.0 / ImuMayhony.sampleFreq));
    gz *= (0.5 * (1.0 / ImuMayhony.sampleFreq));
    qa = ImuMayhony.q0;
    qb = ImuMayhony.q1;
    qc = ImuMayhony.q2;
    ImuMayhony.q0 += (-qb*gx - qc*gy - ImuMayhony.q3*gz);
    ImuMayhony.q1 += (qa*gx + qc*gz - ImuMayhony.q3*gy);
    ImuMayhony.q2 += (qa*gy - qb*gz + ImuMayhony.q3*gx);
    ImuMayhony.q3 += (qa*gz + qb*gy - qc*gx);

    // Normalize quaternion
    recipNorm = invSqrt(ImuMayhony.q0*ImuMayhony.q0 + ImuMayhony.q1*ImuMayhony.q1 + ImuMayhony.q2*ImuMayhony.q2 + ImuMayhony.q3*ImuMayhony.q3);
    ImuMayhony.q0 *= recipNorm;
    ImuMayhony.q1 *= recipNorm;
    ImuMayhony.q2 *= recipNorm;
    ImuMayhony.q3 *= recipNorm;
}

/**
* Returns the yaw pitch and roll angles, respectively defined as the angles in radians between
* the Earth North and the IMU Z axis (yaw), the Earth ground plane and the IMU Y axis (pitch)
* and the Earth ground plane and the IMU X axis (roll).
*
* Returns Yaw, Pitch and Roll angles in radians
**/
void Update(float gx, float gy, float gz, float ax, float ay, float az, float mx, float my, float mz, float *yaw, float *pitch, float *roll)
{
    float64    gravx, gravy, gravz; // estimated gravity direction

    ImuMayhony.sampleFreq = 1.0 / ((float)(2 - ImuMayhony.lastUpdate) / 1000000000.0); // nanoseconds to fractions of a second
    ImuMayhony.lastUpdate = 2;
    ahrsUpdate(gx, gy, gz, ax, ay, az, mx, my, mz);
    gravx  = (float64)(2.0 * (ImuMayhony.q1 * ImuMayhony.q3 - ImuMayhony.q0 * ImuMayhony.q2));
    gravy  = (float64)(2.0 * (ImuMayhony.q0 * ImuMayhony.q1 + ImuMayhony.q2 * ImuMayhony.q3));
    gravz  = (float64)(ImuMayhony.q0 * ImuMayhony.q0 - ImuMayhony.q1 * ImuMayhony.q1 - ImuMayhony.q2 * ImuMayhony.q2 + ImuMayhony.q3 * ImuMayhony.q3);
    *yaw   = (float)(atan2(float64(2.0 * ImuMayhony.q1 * ImuMayhony.q2 - 2.0 * ImuMayhony.q0 * ImuMayhony.q3), float64(2.0 * ImuMayhony.q0 * ImuMayhony.q0 + 2.0 * ImuMayhony.q1 * ImuMayhony.q1 - 1.0)));
    *pitch = (float)(atan(gravx / sqrt(gravy * gravy + gravz * gravz)));
    *roll  = (float)(atan(gravy / sqrt(gravx * gravx + gravz * gravz)));
} 






void calc(C_VAL *curr, C_VAL *last, MOTORS *m)
{  

  float   d2r = M_PI  / 180.0;   // Used to convert degrees to radians 
	float   r2d = 180.0 / M_PI;    // Used to convert radians to degrees 

	yaw, pitch, roll = Update(sData.When, sData.Gx * d2r, sData.Gy * d2r, sData.Gz * d2r, sData.Ax, sData.Ay, sData.Az, sData.Mx, sData.My, sData.Mz) 
	

  // PID ?
	

	fmt.Printf("%d YPR(%10.5f, %10.5f, %10.5f)\n", cnt, yaw*r2d, pitch*r2d, roll*r2d)
	fmt.Printf("%d CMD(%d, %d, %d, %d, %d, %d\n", cData.Yaw, cData.Pitch, cData.Roll, cData.Aux1, cData.Aux2) 


}