// from  Crazyflie Firmware
#include "calc.h"
#include <math.h>
#include <stdint.h>

#define M_PI                  3.14159265358979323846

#define PID_ROLL_RATE_KP                    70.0
#define PID_ROLL_RATE_KI                     0.0
#define PID_ROLL_RATE_KD                     0.0
#define PID_ROLL_RATE_INTEGRATION_LIMIT    100.0

#define PID_PITCH_RATE_KP                   70.0
#define PID_PITCH_RATE_KI                    0.0
#define PID_PITCH_RATE_KD                    0.0
#define PID_PITCH_RATE_INTEGRATION_LIMIT   100.0

#define PID_YAW_RATE_KP                     50.0
#define PID_YAW_RATE_KI                     25.0
#define PID_YAW_RATE_KD                      0.0
#define PID_YAW_RATE_INTEGRATION_LIMIT     100.0

#define PID_ROLL_KP                          3.5
#define PID_ROLL_KI                          2.0
#define PID_ROLL_KD                          0.0
#define PID_ROLL_INTEGRATION_LIMIT          20.0

#define PID_PITCH_KP                         3.5
#define PID_PITCH_KI                         2.0
#define PID_PITCH_KD                         0.0
#define PID_PITCH_INTEGRATION_LIMIT         20.0

#define PID_YAW_KP                           0.0
#define PID_YAW_KI                           0.0
#define PID_YAW_KD                           0.0
#define PID_YAW_INTEGRATION_LIMIT          360.0

#define DEFAULT_PID_INTEGRATION_LIMIT     1000.0


//Fancier version
#define TRUNCATE_SINT16(out, in) (out = (in<INT16_MIN)?INT16_MIN:((in>INT16_MAX)?INT16_MAX:in) )
//Better semantic
#define SATURATE_SINT16(in) ( (in<INT16_MIN)?INT16_MIN:((in>INT16_MAX)?INT16_MAX:in) )

typedef struct
{
  float desired;      //< set point
  float error;        //< error
  float prevError;    //< previous error
  float integ;        //< integral
  float deriv;        //< derivative
  float kp;           //< proportional gain
  float ki;           //< integral gain
  float kd;           //< derivative gain
  float outP;         //< proportional output (debugging)
  float outI;         //< integral output (debugging)
  float outD;         //< derivative output (debugging)
  float iLimit;       //< integral limit
} PidObject;

PidObject pidRollRate;
PidObject pidPitchRate;
PidObject pidYawRate;
PidObject pidRoll;
PidObject pidPitch;
PidObject pidYaw;

S16 rollOutput;
S16 pitchOutput;
S16 yawOutput;

static BOOLEAN isInit = FALSE;



//#define MADWICK_QUATERNION_IMU
#define FUSION_UPDATE_DT    0.02

#ifdef MADWICK_QUATERNION_IMU
 #define BETA_DEF     0.01f          // 2 * proportional gain
#else // MAHONY_QUATERNION_IMU
 #define TWO_KP_DEF  (2.0f * 0.4f)   // 2 * proportional gain
 #define TWO_KI_DEF  (2.0f * 0.001f) // 2 * integral gain
#endif

#ifdef MADWICK_QUATERNION_IMU
  float beta = BETA_DEF;             // 2 * proportional gain (Kp)
#else // MAHONY_QUATERNION_IMU
  float twoKp = TWO_KP_DEF;          // 2 * proportional gain (Kp)
  float twoKi = TWO_KI_DEF;          // 2 * integral gain (Ki)
  float integralFBx = 0.0f;
  float integralFBy = 0.0f;
  float integralFBz = 0.0f;          // integral error terms scaled by Ki
#endif

float q0 = 1.0f;
float q1 = 0.0f;
float q2 = 0.0f;
float q3 = 0.0f;                     // quaternion of sensor frame relative to auxiliary frame

float rollRateDesired;
float pitchRateDesired;
float yawRateDesired;

U16  actuatorThrust;
S16  actuatorRoll;
S16  actuatorPitch;
S16  actuatorYaw;


void pidInit(PidObject* pid, const float desired, const float kp,
             const float ki, const float kd)
{
  pid->error = 0;
  pid->prevError = 0;
  pid->integ = 0;
  pid->deriv = 0;
  pid->desired = desired;
  pid->kp = kp;
  pid->ki = ki;
  pid->kd = kd;
  
  pid->iLimit = DEFAULT_PID_INTEGRATION_LIMIT;
}

float pidUpdate(PidObject* pid, const float measured, const BOOLEAN updateError)
{
  float output;

  if (updateError)
  {
    pid->error = pid->desired - measured;
  }

  pid->integ += pid->error * FUSION_UPDATE_DT;
  if (pid->integ > pid->iLimit)
  {
    pid->integ = pid->iLimit;
  }
  else if (pid->integ < -pid->iLimit)
  {
    pid->integ = -pid->iLimit;
  }

  pid->deriv = (pid->error - pid->prevError) / FUSION_UPDATE_DT;

  pid->outP = pid->kp * pid->error;
  pid->outI = pid->ki * pid->integ;
  pid->outD = pid->kd * pid->deriv;

  output = (pid->kp * pid->error) +
           (pid->ki * pid->integ) +
           (pid->kd * pid->deriv);

  pid->prevError = pid->error;

  return output;
}

void controllerCorrectRatePID(
       float rollRateActual, float pitchRateActual, float yawRateActual,
       float rollRateDesired, float pitchRateDesired, float yawRateDesired)
{
  pidRollRate.desired = rollRateDesired;
  TRUNCATE_SINT16(rollOutput, pidUpdate(&pidRollRate, rollRateActual, TRUE));

  pidPitchRate.desired = pitchRateDesired;
  TRUNCATE_SINT16(pitchOutput, pidUpdate(&pidPitchRate, pitchRateActual, TRUE));

  pidYawRate.desired = yawRateDesired;
  TRUNCATE_SINT16(yawOutput, pidUpdate(&pidYawRate, yawRateActual, TRUE));
}

void controllerCorrectAttitudePID(
       float eulerRollActual, float eulerPitchActual, float eulerYawActual,
       float eulerRollDesired, float eulerPitchDesired, float eulerYawDesired,
       float* rollRateDesired, float* pitchRateDesired, float* yawRateDesired)
{
  float yawError;

  pidRoll.desired = eulerRollDesired;
  *rollRateDesired = pidUpdate(&pidRoll, eulerRollActual, TRUE);

  // Update PID for pitch axis
  pidPitch.desired = eulerPitchDesired;
  *pitchRateDesired = pidUpdate(&pidPitch, eulerPitchActual, TRUE);

  // Update PID for yaw axis
  yawError = eulerYawDesired - eulerYawActual;
  if (yawError > 180.0)
    yawError -= 360.0;
  else if (yawError < -180.0)
    yawError += 360.0;
  pidYaw.error = yawError;
  *yawRateDesired = pidUpdate(&pidYaw, eulerYawActual, FALSE);
}

void controllerGetActuatorOutput(S16* roll, S16* pitch, S16* yaw)
{
  *roll = rollOutput;
  *pitch = pitchOutput;
  *yaw = yawOutput;
}

float invSqrt(float x)
{
  float halfx = 0.5f * x;
  float y = x;
  long i = *(long*)&y;
  i = 0x5f3759df - (i>>1);
  y = *(float*)&i;
  y = y * (1.5f - (halfx * y * y));
  return y;
}

#ifdef MADWICK_QUATERNION_IMU
// Implementation of Madgwick's IMU and AHRS algorithms.
// See: http://www.x-io.co.uk/open-source-ahrs-with-x-imu
//
// Date     Author          Notes
// 29/09/2011 SOH Madgwick    Initial release
// 02/10/2011 SOH Madgwick  Optimised for reduced CPU load
void sensfusion6UpdateQ(float gx, float gy, float gz, float ax, float ay, float az, float dt)
{
  float recipNorm;
  float s0, s1, s2, s3;
  float qDot1, qDot2, qDot3, qDot4;
  float _2q0, _2q1, _2q2, _2q3, _4q0, _4q1, _4q2 ,_8q1, _8q2, q0q0, q1q1, q2q2, q3q3;

  // Rate of change of quaternion from gyroscope
  qDot1 = 0.5f * (-q1 * gx - q2 * gy - q3 * gz);
  qDot2 = 0.5f * (q0 * gx + q2 * gz - q3 * gy);
  qDot3 = 0.5f * (q0 * gy - q1 * gz + q3 * gx);
  qDot4 = 0.5f * (q0 * gz + q1 * gy - q2 * gx);

  // Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
  if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f)))
  {
    // Normalise accelerometer measurement
    recipNorm = invSqrt(ax * ax + ay * ay + az * az);
    ax *= recipNorm;
    ay *= recipNorm;
    az *= recipNorm;

    // Auxiliary variables to avoid repeated arithmetic
    _2q0 = 2.0f * q0;
    _2q1 = 2.0f * q1;
    _2q2 = 2.0f * q2;
    _2q3 = 2.0f * q3;
    _4q0 = 4.0f * q0;
    _4q1 = 4.0f * q1;
    _4q2 = 4.0f * q2;
    _8q1 = 8.0f * q1;
    _8q2 = 8.0f * q2;
    q0q0 = q0 * q0;
    q1q1 = q1 * q1;
    q2q2 = q2 * q2;
    q3q3 = q3 * q3;

    // Gradient decent algorithm corrective step
    s0 = _4q0 * q2q2 + _2q2 * ax + _4q0 * q1q1 - _2q1 * ay;
    s1 = _4q1 * q3q3 - _2q3 * ax + 4.0f * q0q0 * q1 - _2q0 * ay - _4q1 + _8q1 * q1q1 + _8q1 * q2q2 + _4q1 * az;
    s2 = 4.0f * q0q0 * q2 + _2q0 * ax + _4q2 * q3q3 - _2q3 * ay - _4q2 + _8q2 * q1q1 + _8q2 * q2q2 + _4q2 * az;
    s3 = 4.0f * q1q1 * q3 - _2q1 * ax + 4.0f * q2q2 * q3 - _2q2 * ay;
    recipNorm = invSqrt(s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
    s0 *= recipNorm;
    s1 *= recipNorm;
    s2 *= recipNorm;
    s3 *= recipNorm;

    // Apply feedback step
    qDot1 -= beta * s0;
    qDot2 -= beta * s1;
    qDot3 -= beta * s2;
    qDot4 -= beta * s3;
  }

  // Integrate rate of change of quaternion to yield quaternion
  q0 += qDot1 * dt;
  q1 += qDot2 * dt;
  q2 += qDot3 * dt;
  q3 += qDot4 * dt;

  // Normalise quaternion
  recipNorm = invSqrt(q0*q0 + q1*q1 + q2*q2 + q3*q3);
  q0 *= recipNorm;
  q1 *= recipNorm;
  q2 *= recipNorm;
  q3 *= recipNorm;
}
#else // MAHONY_QUATERNION_IMU
// Madgwick's implementation of Mayhony's AHRS algorithm.
// See: http://www.x-io.co.uk/open-source-ahrs-with-x-imu
//
// Date     Author      Notes
// 29/09/2011 SOH Madgwick    Initial release
// 02/10/2011 SOH Madgwick  Optimised for reduced CPU load
void sensfusion6UpdateQ(float gx, float gy, float gz, float ax, float ay, float az, float dt)
{
  float recipNorm;
  float halfvx, halfvy, halfvz;
  float halfex, halfey, halfez;
  float qa, qb, qc;

  gx = gx * M_PI / 180;
  gy = gy * M_PI / 180;
  gz = gz * M_PI / 180;

  // Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
  if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f)))
  {
    // Normalise accelerometer measurement
    recipNorm = invSqrt(ax * ax + ay * ay + az * az);
    ax *= recipNorm;
    ay *= recipNorm;
    az *= recipNorm;

    // Estimated direction of gravity and vector perpendicular to magnetic flux
    halfvx = q1 * q3 - q0 * q2;
    halfvy = q0 * q1 + q2 * q3;
    halfvz = q0 * q0 - 0.5f + q3 * q3;

    // Error is sum of cross product between estimated and measured direction of gravity
    halfex = (ay * halfvz - az * halfvy);
    halfey = (az * halfvx - ax * halfvz);
    halfez = (ax * halfvy - ay * halfvx);

    // Compute and apply integral feedback if enabled
    if(twoKi > 0.0f)
    {
      integralFBx += twoKi * halfex * dt;  // integral error scaled by Ki
      integralFBy += twoKi * halfey * dt;
      integralFBz += twoKi * halfez * dt;
      gx += integralFBx;  // apply integral feedback
      gy += integralFBy;
      gz += integralFBz;
    }
    else
    {
      integralFBx = 0.0f; // prevent integral windup
      integralFBy = 0.0f;
      integralFBz = 0.0f;
    }

    // Apply proportional feedback
    gx += twoKp * halfex;
    gy += twoKp * halfey;
    gz += twoKp * halfez;
  }

  // Integrate rate of change of quaternion
  gx *= (0.5f * dt);   // pre-multiply common factors
  gy *= (0.5f * dt);
  gz *= (0.5f * dt);
  qa = q0;
  qb = q1;
  qc = q2;
  q0 += (-qb * gx - qc * gy - q3 * gz);
  q1 += (qa * gx + qc * gz - q3 * gy);
  q2 += (qa * gy - qb * gz + q3 * gx);
  q3 += (qa * gz + qb * gy - qc * gx);

  // Normalise quaternion
  recipNorm = invSqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
  q0 *= recipNorm;
  q1 *= recipNorm;
  q2 *= recipNorm;
  q3 *= recipNorm;
}
#endif

void sensfusion6GetEulerRPY(float* roll, float* pitch, float* yaw)
{
  float gx, gy, gz; // estimated gravity direction

  gx = 2 * (q1*q3 - q0*q2);
  gy = 2 * (q0*q1 + q2*q3);
  gz = q0*q0 - q1*q1 - q2*q2 + q3*q3;

  *yaw = atan2(2*q1*q2 - 2*q0*q3, 2*q0*q0 + 2*q1*q1 - 1) * 180 / M_PI;
  *pitch = atan(gx / sqrt(gy*gy + gz*gz)) * 180 / M_PI;
  *roll = atan(gy / sqrt(gx*gx + gz*gz)) * 180 / M_PI;
}

void controllerInit()
{
  if(isInit == TRUE) return;
  
  pidInit(&pidRollRate, 0, PID_ROLL_RATE_KP, PID_ROLL_RATE_KI, PID_ROLL_RATE_KD);
  pidInit(&pidPitchRate, 0, PID_PITCH_RATE_KP, PID_PITCH_RATE_KI, PID_PITCH_RATE_KD);
  pidInit(&pidYawRate, 0, PID_YAW_RATE_KP, PID_YAW_RATE_KI, PID_YAW_RATE_KD);
  pidRollRate.iLimit  = PID_ROLL_RATE_INTEGRATION_LIMIT;
  pidPitchRate.iLimit = PID_PITCH_RATE_INTEGRATION_LIMIT;
  pidYawRate.iLimit   = PID_YAW_RATE_INTEGRATION_LIMIT;

  pidInit(&pidRoll, 0, PID_ROLL_KP, PID_ROLL_KI, PID_ROLL_KD);
  pidInit(&pidPitch, 0, PID_PITCH_KP, PID_PITCH_KI, PID_PITCH_KD);
  pidInit(&pidYaw, 0, PID_YAW_KP, PID_YAW_KI, PID_YAW_KD);
  pidRoll.iLimit  = PID_ROLL_INTEGRATION_LIMIT;
  pidPitch.iLimit = PID_PITCH_INTEGRATION_LIMIT;
  pidYaw.iLimit   = PID_YAW_INTEGRATION_LIMIT;
  
  isInit = TRUE;
}





/**********************************************************************************
*            calc with MotionTracking & user-command for motor control            *
*                                                                                 *
* input:                                                                          *
*    curr   =   [in]     - new or unchanged values from MPU & user control as FP  *
*                normalized accel (0.001g) & gyro (0.01/s)                       *
*                mag (unused)                                                     *
*                normalized user control (+/- 350 and 0-700)                      *
*                 acc_x,     acc_y,    acc_z                                      *
*                 gyr_pitch, gyr_roll, gyr_yaw                                    *
*                 mag_x,     mag_y,    mag_z                                      *
*                 usr_pitch, usr_roll, usr_yaw, usr_z                             *
*    last   =   [in/out] - last calculated values (in here)                       *
*                same elements like "curr", use is implementation specific        *
*    m      =   [out]    - motor[0..3] values between 0...1000                    *
**********************************************************************************/
void calc(C_VAL *curr, C_VAL *last, MOTORS *m)
{
    controllerInit();
	
    sensfusion6UpdateQ(curr->gyr_pitch, curr->gyr_roll, curr->gyr_yaw, curr->acc_x, curr->acc_y, curr->acc_z, FUSION_UPDATE_DT);
    sensfusion6GetEulerRPY(&last->gyr_roll, &last->gyr_pitch, &last->gyr_yaw);
    controllerCorrectAttitudePID(last->gyr_roll, last->gyr_pitch, last->gyr_yaw,
                                 curr->usr_roll, curr->usr_pitch, -curr->usr_yaw,
                                 &rollRateDesired, &pitchRateDesired, &yawRateDesired);
	
    yawRateDesired = -curr->usr_yaw;
    controllerCorrectRatePID(curr->gyr_pitch, curr->gyr_roll, curr->gyr_yaw,
                             rollRateDesired, pitchRateDesired, yawRateDesired);

    controllerGetActuatorOutput(&actuatorRoll, &actuatorPitch, &actuatorYaw);

    // Calculate motor commands
    actuatorPitch = actuatorPitch >> 1; actuatorRoll = actuatorRoll >> 1; 
    m->motor[0] = (U16)(curr->usr_z - actuatorPitch - actuatorRoll - actuatorYaw); if(m->motor[0] > 1000) m->motor[0] = 1000;
    m->motor[1] = (U16)(curr->usr_z + actuatorPitch - actuatorRoll + actuatorYaw); if(m->motor[1] > 1000) m->motor[1] = 1000;
    m->motor[2] = (U16)(curr->usr_z - actuatorPitch + actuatorRoll + actuatorYaw); if(m->motor[2] > 1000) m->motor[2] = 1000;
    m->motor[3] = (U16)(curr->usr_z + actuatorPitch + actuatorRoll - actuatorYaw); if(m->motor[3] > 1000) m->motor[3] = 1000;
	
}



